/*
 * Copyright (c) 2007, 2008 Sandvine Incorporated.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ctype.h>
#include <dialog.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void
init_dialog()
{
}

void
end_dialog()
{
}

void
dialog_clear()
{
}

void
dialog_clear_norefresh()
{
}

static void
print_title(const char *title)
{
	int i;
	printf("%s\n", title);
	for (i = 0; i < strlen(title); ++i)
		putchar('=');
	printf("\n");
}

/*
     To use the newer interface,
     cnt should be a negative integer representing the number of
     dialogMenuItem structures pointed to by it (which should be of type
     dialogMenuItem *), one structure per item.  In the new interface, the
     result variable is used as a simple boolean (not a pointer) and should be
     NULL if it only points to menu items and the default OK and Cancel but-
     tons are desired.  If result is non-NULL, then it is actually expected to
     point 2 locations past the start of the menu item list.  it is then
     expected to point to an item representing the Cancel button, from which
     the prompt and fire actions are used to override the default behavior,
     and it to the same for the OK button.

     Using either API behavior, the ch and sc values may be passed in to pre-
     serve current item selection and scroll position values across calls.
*/

/*
typedef struct _dmenu_item {
char *prompt;
char *title;
int (*checked)(struct _dmenu_item *self);
int (*fire)(struct _dmenu_item *self);
int (*selected)(struct _dmenu_item *self, int is_selected);
void *data;
char lbra, mark, rbra;
long aux;
} dialogMenuItem;
*/

static void
read_string(char *buf, size_t len)
{
	fgets(buf, len, stdin);
	buf[len - 1] = '\0';
	strsep(&buf, "\r\n");
}

static int
do_yesno(unsigned char *title, unsigned char *prompt, int default_yes)
{
	char buf[16];

	if (title)
		print_title(title);
	while (1) {
		printf("%s\n(y/n) [%c]> ", prompt, default_yes ? 'y' : 'n');
		read_string(buf, sizeof(buf));
		switch (buf[0]) {
		case '\0':
			return default_yes;
		case 'y':
		case 'Y':
			return 1;
		case 'n':
		case 'N':
			return 0;
		}
		printf ("Invalid response\n");
	}
}

int
dialog_menu(unsigned char *title, unsigned char *prompt, int height,
    int width, int menu_height, int cnt, void *it, unsigned char *result,
    int *ch, int *sc)
{
	char buf[16];
	dialogMenuItem *menu_item;
	int i;
restart_title:
	print_title(title);

	if (cnt >= 0 || result) {
		printf("Old/full %s API not implemented\n", __func__);
		abort();
	}
	menu_item = (dialogMenuItem *)it;
	while (1) {
		for (i=0; i < -cnt; ++i) {
			if (menu_item[i].checked != 0 ||
			    menu_item[i].selected != 0) {
				printf("Full %s API not implemented\n",
				     __func__);
				abort();
			}
			if (i > 0 && i % 20 == 0) {
				char c;
				do {
					printf("More? (y/n) [y] " );
					read_string(buf, sizeof(buf));
					c = toupper(buf[0]);
				} while (c != '\0' && c != 'Y' && c != 'N');
				if (c == 'N')
					break;
			}
			printf("%s. %s\n", menu_item[i].prompt,
			    menu_item[i].title);
		}
		printf("%s\n> ", prompt);
		read_string(buf, sizeof(buf));
		for (i = 0; i < -cnt; ++i) {
			if (strcmp(buf, menu_item[i].prompt) == 0) {
				int rv;
				rv = menu_item[i].fire(&menu_item[i]);
				if (rv & DITEM_LEAVE_MENU)
					return 0;
				else if (rv & DITEM_RECREATE)
					goto restart_title;
			}
		}

		if (toupper(buf[0]) == 'B')
			return 1;
		else if (toupper(buf[0]) == 'Q')
			return -1;
		
		printf("Selection not recognized\n");
		if (do_yesno(NULL, "Return to previous menu?", 0))
			return 1;
	}
	return 0;
}

int
dialog_mesgbox(unsigned char *title, unsigned char *prompt, int height,
    int width)
{
	print_title(title);
	printf("%s\n", prompt);
	return 0;
}

int
dialog_yesno(unsigned char *title, unsigned char *prompt, int height,
    int width)
{
	return !do_yesno(title, prompt, 1);
}

int
dialog_noyes(unsigned char *title, unsigned char *prompt, int height,
    int width)
{
	return !do_yesno(title, prompt, 0);
}
