#!/bin/sh

# setupSources should run on the local Cygwin machine
setupSources()
{
    VIEW=$1
    WORKDIR=$2
    BUILDMACHINE=$3
    ssh $BUILDMACHINE mkdir -p $WORKDIR
    test $? -eq 0 || exit 1
    cd /m/$VIEW/fw-bsd/
    tar -c -z -f - src | ssh $BUILDMACHINE tar -x -z -C $WORKDIR -f -
    test $? -eq 0 || exit 1
    cd /m/$VIEW/fw/
    tar -c -z -f - products/freebsd/src | ssh $BUILDMACHINE tar -x -z -C $WORKDIR -f -
    test $? -eq 0 || exit 1
}

# makeImage should run on the build machine
makeImage()
{
    WORKDIR=$1
    ROOTDIR=$2
    VERSION=$3
    cd $WORKDIR
    mkdir -p $ROOTDIR/usr
    sh src/sv/setupchroot.sh $ROOTDIR
    cp -RPp src $ROOTDIR/usr
    cp -RPp products $ROOTDIR/
    chroot $ROOTDIR /bin/sh << EOF
set -e
trap 'umount /dev || true' 0
mount -t devfs none /dev
cd /usr/src/sv
make TARGETVERSION=$VERSION KERNELS="USBROOT USBROOT-NOSMP" \
    DEFAULTKERNEL=USBROOT INSTALLSCRIPT=usbinstallscript.sh \
    sv_installroot
EOF
}

# makeArchive should run on the build machine
makeArchive()
{
    ROOTDIR=$1
    VERSION=$2
    cd $ROOTDIR
    cat usr/local/libdata/image_packing_list | sed 's/^\///' > image_list
    echo 'usr/local/bin/socket' >> image_list
    tar -c -z -n -f FREEBSD_$VERSION.tgz -T image_list
}

case $1 in
    makeall)
	shift
        VIEW=$1
        WORKDIR=$2
        ROOTDIR=chrootdir
        VERSION=$3
        BUILDMACHINE=$4
	setupSources $VIEW $WORKDIR $BUILDMACHINE
	#ssh $BUILDMACHINE $WORKDIR/src/sv/buildusb.sh buildimage $WORKDIR $ROOTDIR $VERSION
    ;;
    buildimage)
	shift
        WORKDIR=$1
        ROOTDIR=$2
        VERSION=$3
	makeImage $WORKDIR $ROOTDIR $VERSION
	makeArchive $ROOTDIR $VERSION
    ;;
    *)
	echo "Usage: builddiskless.sh makeall (on Cygwin)"
	echo "       builddiskless.sh buildimage (on build machine)"
	exit 1
    ;;
esac
