#!/bin/sh

# Create a virtual package containing the files specified by packing_file,
# excluding those listed in exclude_file (passed to egrep -f)

usage()
{
    cat <<EOF
$(basename $0) packing_file exclude_file packagename
EOF
}

if [ $# -lt 3 ]; then
    usage
    exit 2
fi

packlist=$1
exclude=$2
pkgname=$3

if [ ! -f $packlist -o ! -f $exclude ]; then
    echo "$packlist or $exclude not found"
    usage
    exit 2
fi

# Output a +CONTENTS style list of
# filename
# @comment MD5:<hash>
getMd5s()
{
    local md5
    for f in $(egrep -v -f $exclude $packlist); do
        if [ -f /$f ]; then
            echo $f | sed 's/^\///'
            if [ -L /$f ]; then
                # Symlinks have the link dest hashed.  The tr command strips
                # the trailing newline.
                md5=$(readlink /$f | tr -d '\012' | md5)
            else
                md5=$(cat /$f | md5)
            fi
            echo "@comment MD5:$md5"
        fi
    done
}

if [ "$(id -u)" != 0 ]; then
#    echo "Invoking $0 as root"
#    exec sudo $0 $*
#    echo "Odd, this returned?"
    echo "Must be run as root."
    exit 1
fi

pkgpath=/var/db/pkg/$pkgname
mkdir -p $pkgpath
# Make sure it's empty, so we don't include this package in the contents file
rm -f $pkgpath/*
cat <<EOF > $pkgpath/+COMMENT
Sandvine FreeBSD image pseudo-package
EOF
cat <<EOF > $pkgpath/+DESC
Sandvine FreeBSD image virtual package.  This package is a placeholder, so that
you may use pkg_info -g to list files that have been modified.
EOF
cat <<EOF > $pkgpath/+CONTENTS
@comment PKG_FORMAT_REVISION:1.1
@comment ORIGIN:sandvine/$pkgname
@name $pkgname
@cwd /
$(getMd5s)
EOF
cat <<EOF > $pkgpath/+DEINSTALL
cat <<END >&2
This is a virtual package for tracking changes to files included with the
original stock image and/or upgrade packages.  It may not be deleted.
END
exit 64
EOF
