DISK_SECTS=63	# per track
DISK_HEADS=255

# Determine the primary disk (the one we will image)
find_disk()
{
	DISK_DEV=
	for disk in `sysctl -n kern.disks`; do
		case $disk in
			da0|aacd0)
				if [ -n "$DISK_DEV" ]; then
					echo "Cannot determine primary disk">&2
					exit 1
				fi
				DISK_DEV="$disk"
		esac
	done
}

# Partition config
fdisk_setup()
{
	DISK_SIZE=`fdisk -t -I $DISK_DEV | awk '/size/ {print $4}'`
	PART1_SIZE=1638400 # 800 MB
	PART3_SIZE=1024000 # 500 MB for Rembo cache
	PART2_SIZE=`expr $DISK_SIZE - $PART1_SIZE - $PART3_SIZE`
}

# Output an fdisk(1) config file
fdisk_config()
{
	echo $DISK_SIZE $DISK_SECTS $DISK_HEADS \
	    $PART1_SIZE $PART2_SIZE $PART3_SIZE |
		awk '
		{
			# size of cylinder in sectors
			cs = $2 * $3

			# number of full cylinders on media
			cyl = int ($1 / cs)

			printf "# %d sectors x %d heads\n", $2, $3
			printf "# %d sectors per cylinder\n", cs
			printf "# %d cylinders on disk (%d blocks)\n", cyl, $1

			# output fdisk geometry spec, truncate cyls to 1023
			if (cyl <= 1023)
				print "g c" cyl " h" $3 " s" $2
			else
				print "g c" 1023 " h" $3 " s" $2

			p1sl = int (($4 + cs - 1) / cs)
			p2sl = int ($5 / cs)
			p3sl = int (($6 + cs - 1) / cs)

			print "p 1 165 " $2, p1sl * cs - $2
			print "p 2 165 " p1sl * cs, p2sl * cs
			print "p 3 0 0 0"
			print "p 4 0 0 0"
			print "a 1"
		}'
}

# Wipe a disk device
# $1 - disk device
# $2 - size in 512 byte sectors
zero_disk()
{
	local bytes size offset
	bytes=`expr $2 \* 512`
	size=1048576 # 1MB
	offset=0

	while [ $size -ge 512 -a $bytes -gt 0 ]; do
		count=`expr $bytes / $size`
		if [ $count -gt 0 ]; then
			dd if=/dev/zero of=$1 \
			    oseek=$offset count=$count bs=$size
		fi
		bytes=`expr $bytes - \( $count \* $size \)`
		offset=`expr \( $offset + $count \) \* 2`
		size=`expr $size / 2`
	done
}

find_disk
fdisk_setup
echo "Disk size = $DISK_SIZE (`expr $DISK_SIZE / 2 / 1024 / 1024`G)"

fdisk_config | fdisk -i -f - $DISK_DEV
zero_disk /dev/${DISK_DEV}s1 $PART1_SIZE
bsdlabel -w -B ${DISK_DEV}s1
newfs -U /dev/${DISK_DEV}s1a

mkdir -p /mnt/new_root
mount /dev/${DISK_DEV}s1a /mnt/new_root
