#!/bin/sh

# Create generic ${TOPDIR}/etc/fstab for the disk image.  The generic fstab
# contains a / entry for each possible slice we might boot from, and this same
# list is inserted into the loader's vfs.root.mountfrom.  The kernel will try
# each in turn, and when the device comes up the disklabel script removes all
# but the one that was actually mounted.  This used to be handled by
# prep_image.

TOPDIR=/
while getopts "d:" opt; do
case $opt in
	d)
		TOPDIR="$OPTARG"
		;;
esac
done

for dir in ${TOPDIR} ${TOPDIR}/etc ${TOPDIR}/boot; do
    if [ ! -d $dir ]; then
	echo "Directory <$dir> does not exist." >&2
	exit 1
    fi
done

cat <<EOF >${TOPDIR}/etc/fstab
# Device	Mountpoint		FStype	Options		     Dump Pass#
/dev/ad0s1a	/			ufs	rw			1 1
/dev/da0s1a	/			ufs	rw			1 1
/dev/aacd0s1a	/			ufs	rw			1 1
/dev/amrd0s1a	/			ufs	rw			1 1
/dev/twed0s1a	/			ufs	rw			1 1
/dev/acd0c	/cdrom			cd9660	ro,noauto		0 0
proc		/proc			procfs	rw			0 0
EOF

# This whole thing is gross.  vfs.root.mountfrom will take a space-separated
# list and try each one in turn.  One of them will succeed, and then the
# this line can be removed from /boot/loader.conf by disklabel (because then
# fstab will have only one / entry).
devlist=""
while read dev mount fstype rest; do
    if [ "$mount" = / ]; then
	if [ -n "$devlist" ]; then
	    devlist="$devlist "
	fi
	devlist="$devlist$fstype:$dev"
    fi
done < ${TOPDIR}/etc/fstab
cat <<EOF >>${TOPDIR}/boot/loader.conf
vfs.root.mountfrom="$devlist" # setup_fstab.sh
EOF

