#include "sv_group.h"

#include <syslog.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Search and replace "user"'s entry with new entry "user	groups". */
void ReplaceSvGroupsEntry(int fd, char *user, char *groups, char *buffer __attribute__ ((unused)), unsigned int bufSize __attribute__ ((unused)) )
{
    char *file_contents, *newfile_contents;
    int offset, newoffset, allowable_groups_entry_replaced;
    char buf_user[MAX_USERNAME_LEN+1];
    char buf_groups[CMD_BUF_SIZE+1];

    allowable_groups_entry_replaced = 0;
    newoffset = 0;

    /* Read the old groups file, entry by entry.  If the entry's user is the
     * user we wish to replace, do so and store it in newfile_contents.
     * Otherwise, store that entry in newfile_contents.
     */
    if ((file_contents = ReadFile(fd)) != NULL)
    {
	/* Allocate enough space for the new file contents.  We need at a
	 * maximum, the size of the old file contents plus the size of the
	 * username and groups plus 2 bytes for the tab and newline.
	 */
	if ((newfile_contents = (char *)malloc(strlen(file_contents)+strlen(user)+strlen(groups)+3)) == NULL)
	{
	    syslog(LOG_CRIT, "System out of resources : malloc.\n");
	    free(file_contents);
	    return;
	}
	offset = 0;
        allowable_groups_entry_replaced = 0;
	while ((offset = ReadSvGroupsEntry(file_contents, offset, buf_user, buf_groups)) != -1)
	{
	    if (!strcmp(user, buf_user))
	    {
		sprintf(newfile_contents+newoffset, "%s\t%s\n", user, groups);
		newoffset += strlen(user) + strlen(groups) + 2;
		allowable_groups_entry_replaced = 1;
	    }
	    else
	    {
		sprintf(newfile_contents+newoffset, "%s\t%s\n", buf_user, buf_groups);
		newoffset += strlen(buf_user) + strlen(buf_groups) + 2;
	    }
	}
	free(file_contents);
    }
    else
    {
	/* Allocate enough space for the new file contents.  We need at a
	 * maximum, the size of the old file contents plus the size of the
	 * username and groups plus 2 bytes for the tab and newline.
	 */
	if ((newfile_contents = (char *)malloc(strlen(user)+strlen(groups)+3)) == NULL)
	{
	    syslog(LOG_CRIT, "System out of resources : malloc.\n");
	    return;
	}
    }
    if (!allowable_groups_entry_replaced)
    {
	sprintf(newfile_contents+newoffset, "%s\t%s\n", user, groups);
	newoffset += strlen(user) + strlen(groups) + 2;
    }

    /* Move the contents of the temporary groups file to the actual groups file. */
    ftruncate(fd, 0);
    if (lseek(fd, 0, SEEK_SET) == -1)
    {
	syslog(LOG_CRIT, "Could not seek to beginning of file.\n");
	free(newfile_contents);
	return;
    }
    if (write(fd, newfile_contents, newoffset) != newoffset)
    {
	syslog(LOG_CRIT, "Final move of entire contents to %s failed.\n", SANDVINE_GROUPS_FILE);
    }
    free(newfile_contents);
}

/*
 * Search for user and return its groups entry. Assumes groups has enough buffer
 * space.  Return 0 if found and 1 otherwise.
 */
int
SearchSvGroupsEntry(int fd, char *user, char *groups)
{
    char *file_contents;
    int offset;
    char buf_user[MAX_USERNAME_LEN+1];

    /* Read the old groups file, entry by entry.  If the entry's user is the
     * user passed in, break and return 0.
     */
    if ((file_contents = ReadFile(fd)) != NULL)
    {
        offset = 0;
	while ((offset = ReadSvGroupsEntry(file_contents, offset, buf_user, groups)) != -1)
	{
	    if (!strcmp(user, buf_user))
	    {
		break;
	    }
	}
        free(file_contents);
	if (offset == -1)
	{
	    return 1;
	}
	return 0;
    }
    return 1;
}

/* Retrieve the user and group from buffer.  Assumes user and groups has
 *  sufficient buffer space.
 */
int
ReadSvGroupsEntry(char *buffer, int offset, char *user, char *groups)
{
    char buf[CMD_BUF_SIZE];
    char *s, *ss;
    if ((s = (char *)strchr(buffer+offset, '\n')) == NULL)
    {
	return -1;
    }
    strncpy(buf, buffer+offset, s-(buffer+offset));
    buf[s-(buffer+offset)] = '\0';
    if ((ss = (char *)strchr(buf, '\t')) == NULL)
    {
	syslog(LOG_CRIT, "Entry %s missing tab character.\n", buf);
	return -1;
    }
    strncpy(user, buf, ss-buf);
    user[ss-buf] = '\0';
    strcpy(groups, ss+1);
    groups[strlen(ss+1)] = '\0';
    offset = s + 1 - buffer;
    return offset;
}

/* Read a complete file.  Make sure you free it after using it! */
char *
ReadFile(int fd)
{
    char *buf;
    unsigned int file_size;
    struct stat file_stat;
    if (!fstat(fd, &file_stat))
    {
	file_size = file_stat.st_size;
	if (!file_size)
	{
	    return NULL;
	}
	if ((buf = (char *)malloc(file_size+1)) == NULL)
	{
	    syslog(LOG_CRIT, "System out of resources : malloc.\n");
	    return NULL;
	}
        if (lseek(fd, 0, SEEK_SET) == -1)
	{
	    syslog(LOG_CRIT, "Could not seek to beginning of file.\n");
	    return NULL;
	}
	if (read(fd, (void *)buf, file_size) != (int)file_size)
	{
	    syslog(LOG_CRIT, "Did not read all of %s.\n", SANDVINE_GROUPS_FILE);
	    free(buf);
	    return NULL;
	}
	buf[file_size] = '\0';
    }
    else
    {
	syslog(LOG_CRIT, "Could not get stat of file. Error number %d.\n", errno);
	return NULL;
    }
    return buf;
}
